@@ -7,6 +7,7 @@ import sys
 from asyncio.base_events import _run_until_complete_cb  # type: ignore
 from collections import OrderedDict, deque
 from concurrent.futures import Future
+from contextvars import Context, copy_context
 from dataclasses import dataclass
 from functools import partial, wraps
 from inspect import (
@@ -22,6 +23,8 @@ from typing import (
     Mapping, Optional, Sequence, Set, Tuple, Type, TypeVar, Union, cast)
 from weakref import WeakKeyDictionary
 
+import sniffio
+
 from .. import CapacityLimiterStatistics, EventStatistics, TaskInfo, abc
 from .._core._compat import DeprecatedAsyncContextManager, DeprecatedAwaitable
 from .._core._eventloop import claim_worker_thread, threadlocals
@@ -305,7 +308,7 @@ class CancelScope(BaseCancelScope):
 
         host_task_state.cancel_scope = self._parent_scope
 
-        # Restart the cancellation effort in the nearest directly cancelled parent scope if this
+        # Restart the cancellation effort in the farthest directly cancelled parent scope if this
         # one was shielded
         if self._shield:
             self._deliver_cancellation_to_parent()
@@ -365,12 +368,12 @@ class CancelScope(BaseCancelScope):
             self._cancel_handle = None
 
     def _deliver_cancellation_to_parent(self) -> None:
-        """Start cancellation effort in the nearest directly cancelled parent scope"""
+        """Start cancellation effort in the farthest directly cancelled parent scope"""
         scope = self._parent_scope
+        scope_to_cancel: Optional[CancelScope] = None
         while scope is not None:
             if scope._cancel_called and scope._cancel_handle is None:
-                scope._deliver_cancellation()
-                break
+                scope_to_cancel = scope
 
             # No point in looking beyond any shielded scope
             if scope._shield:
@@ -378,6 +381,9 @@ class CancelScope(BaseCancelScope):
 
             scope = scope._parent_scope
 
+        if scope_to_cancel is not None:
+            scope_to_cancel._deliver_cancellation()
+
     def _parent_cancelled(self) -> bool:
         # Check whether any parent has been cancelled
         cancel_scope = self._parent_scope
@@ -512,8 +518,9 @@ class ExceptionGroup(BaseExceptionGroup):
 
 
 class _AsyncioTaskStatus(abc.TaskStatus):
-    def __init__(self, future: asyncio.Future):
+    def __init__(self, future: asyncio.Future, parent_id: int):
         self._future = future
+        self._parent_id = parent_id
 
     def started(self, value: object = None) -> None:
         try:
@@ -521,6 +528,9 @@ class _AsyncioTaskStatus(abc.TaskStatus):
         except asyncio.InvalidStateError:
             raise RuntimeError("called 'started' twice on the same task status") from None
 
+        task = cast(asyncio.Task, current_task())
+        _task_states[task].parent_id = self._parent_id
+
 
 class TaskGroup(abc.TaskGroup):
     def __init__(self) -> None:
@@ -650,7 +660,11 @@ class TaskGroup(abc.TaskGroup):
 
         kwargs = {}
         if task_status_future:
-            kwargs['task_status'] = _AsyncioTaskStatus(task_status_future)
+            parent_id = id(current_task())
+            kwargs['task_status'] = _AsyncioTaskStatus(task_status_future,
+                                                       id(self.cancel_scope._host_task))
+        else:
+            parent_id = id(self.cancel_scope._host_task)
 
         coro = func(*args, **kwargs)
         if not asyncio.iscoroutine(coro):
@@ -665,7 +679,7 @@ class TaskGroup(abc.TaskGroup):
             task.add_done_callback(task_done)
 
         # Make the spawned task inherit the task group's cancel scope
-        _task_states[task] = TaskState(parent_id=id(current_task()), name=name,
+        _task_states[task] = TaskState(parent_id=parent_id, name=name,
                                        cancel_scope=self.cancel_scope)
         self.cancel_scope._tasks.add(task)
         return task
@@ -698,24 +712,25 @@ _Retval_Queue_Type = Tuple[Optional[T_Retval], Optional[BaseException]]
 
 
 class WorkerThread(Thread):
-    __slots__ = 'root_task', 'workers', 'idle_workers', 'loop', 'queue', 'idle_since'
-
     MAX_IDLE_TIME = 10  # seconds
 
     def __init__(self, root_task: asyncio.Task, workers: Set['WorkerThread'],
-                 idle_workers: Deque['WorkerThread'],):
+                 idle_workers: Deque['WorkerThread']):
         super().__init__(name='AnyIO worker thread')
         self.root_task = root_task
         self.workers = workers
         self.idle_workers = idle_workers
         self.loop = root_task._loop
-        self.queue: Queue[Union[Tuple[Callable, tuple, asyncio.Future], None]] = Queue(2)
+        self.queue: Queue[Union[Tuple[Context, Callable, tuple, asyncio.Future], None]] = Queue(2)
         self.idle_since = current_time()
+        self.stopping = False
 
     def _report_result(self, future: asyncio.Future, result: Any,
                        exc: Optional[BaseException]) -> None:
         self.idle_since = current_time()
-        self.idle_workers.append(self)
+        if not self.stopping:
+            self.idle_workers.append(self)
+
         if not future.cancelled():
             if exc is not None:
                 future.set_exception(exc)
@@ -731,12 +746,12 @@ class WorkerThread(Thread):
                     # Shutdown command received
                     return
 
-                func, args, future = item
+                context, func, args, future = item
                 if not future.cancelled():
                     result = None
                     exception: Optional[BaseException] = None
                     try:
-                        result = func(*args)
+                        result = context.run(func, *args)
                     except BaseException as exc:
                         exception = exc
 
@@ -747,6 +762,7 @@ class WorkerThread(Thread):
                 self.queue.task_done()
 
     def stop(self, f: Optional[asyncio.Task] = None) -> None:
+        self.stopping = True
         self.queue.put_nowait(None)
         self.workers.discard(self)
         try:
@@ -796,7 +812,9 @@ async def run_sync_in_worker_thread(
                     expired_worker.root_task.remove_done_callback(expired_worker.stop)
                     expired_worker.stop()
 
-            worker.queue.put_nowait((func, args, future))
+            context = copy_context()
+            context.run(sniffio.current_async_library_cvar.set, None)
+            worker.queue.put_nowait((context, func, args, future))
             return await future
 
 
@@ -813,7 +831,11 @@ def run_sync_from_thread(func: Callable[..., T_Retval], *args: object,
 
     f: concurrent.futures.Future[T_Retval] = Future()
     loop = loop or threadlocals.loop
-    loop.call_soon_threadsafe(wrapper)
+    if sys.version_info < (3, 7):
+        loop.call_soon_threadsafe(copy_context().run, wrapper)
+    else:
+        loop.call_soon_threadsafe(wrapper)
+
     return f.result()
 
 
@@ -996,14 +1018,14 @@ def setup_process_pool_exit_at_shutdown(workers: Set[Process]) -> None:
 class StreamProtocol(asyncio.Protocol):
     read_queue: Deque[bytes]
     read_event: asyncio.Event
-    write_future: asyncio.Future
+    write_event: asyncio.Event
     exception: Optional[Exception] = None
 
     def connection_made(self, transport: asyncio.BaseTransport) -> None:
         self.read_queue = deque()
         self.read_event = asyncio.Event()
-        self.write_future = asyncio.Future()
-        self.write_future.set_result(None)
+        self.write_event = asyncio.Event()
+        self.write_event.set()
         cast(asyncio.Transport, transport).set_write_buffer_limits(0)
 
     def connection_lost(self, exc: Optional[Exception]) -> None:
@@ -1012,11 +1034,7 @@ class StreamProtocol(asyncio.Protocol):
             self.exception.__cause__ = exc
 
         self.read_event.set()
-        self.write_future = asyncio.Future()
-        if self.exception:
-            self.write_future.set_exception(self.exception)
-        else:
-            self.write_future.set_result(None)
+        self.write_event.set()
 
     def data_received(self, data: bytes) -> None:
         self.read_queue.append(data)
@@ -1027,10 +1045,10 @@ class StreamProtocol(asyncio.Protocol):
         return True
 
     def pause_writing(self) -> None:
-        self.write_future = asyncio.Future()
+        self.write_event = asyncio.Event()
 
     def resume_writing(self) -> None:
-        self.write_future.set_result(None)
+        self.write_event.set()
 
 
 class DatagramProtocol(asyncio.DatagramProtocol):
@@ -1079,6 +1097,7 @@ class SocketStream(abc.SocketStream):
     async def receive(self, max_bytes: int = 65536) -> bytes:
         with self._receive_guard:
             await checkpoint()
+
             if not self._protocol.read_event.is_set() and not self._transport.is_closing():
                 self._transport.resume_reading()
                 await self._protocol.read_event.wait()
@@ -1092,7 +1111,7 @@ class SocketStream(abc.SocketStream):
                 elif self._protocol.exception:
                     raise self._protocol.exception
                 else:
-                    raise EndOfStream
+                    raise EndOfStream from None
 
             if len(chunk) > max_bytes:
                 # Split the oversized chunk
@@ -1109,19 +1128,21 @@ class SocketStream(abc.SocketStream):
     async def send(self, item: bytes) -> None:
         with self._send_guard:
             await checkpoint()
+
+            if self._closed:
+                raise ClosedResourceError
+            elif self._protocol.exception is not None:
+                raise self._protocol.exception
+
             try:
                 self._transport.write(item)
             except RuntimeError as exc:
-                if self._protocol.write_future.exception():
-                    await self._protocol.write_future
-                elif self._closed:
-                    raise ClosedResourceError from None
-                elif self._transport.is_closing():
+                if self._transport.is_closing():
                     raise BrokenResourceError from exc
                 else:
                     raise
 
-            await self._protocol.write_future
+            await self._protocol.write_event.wait()
 
     async def send_eof(self) -> None:
         try:
@@ -1372,6 +1393,7 @@ class UNIXSocketListener(abc.SocketListener):
             while True:
                 try:
                     client_sock, _ = self.__raw_socket.accept()
+                    client_sock.setblocking(False)
                     return UNIXSocketStream(client_sock)
                 except BlockingIOError:
                     f: asyncio.Future = asyncio.Future()
@@ -1842,7 +1864,7 @@ async def wait_all_tasks_blocked() -> None:
 
 
 class TestRunner(abc.TestRunner):
-    def __init__(self, debug: bool = False, use_uvloop: bool = True,
+    def __init__(self, debug: bool = False, use_uvloop: bool = False,
                  policy: Optional[asyncio.AbstractEventLoopPolicy] = None):
         _maybe_set_event_loop_policy(policy, use_uvloop)
         self._loop = asyncio.new_event_loop()
