@@ -2,16 +2,20 @@ import array
 import math
 import socket
 from concurrent.futures import Future
+from contextvars import copy_context
 from dataclasses import dataclass
 from functools import partial
 from io import IOBase
+from os import PathLike
 from types import TracebackType
 from typing import (
-    Any, Awaitable, Callable, Collection, Coroutine, Dict, Generic, List, NoReturn, Optional, Set,
-    Tuple, Type, TypeVar, Union)
+    Any, Awaitable, Callable, Collection, ContextManager, Coroutine, Deque, Dict, Generic, List,
+    Mapping, NoReturn, Optional, Sequence, Set, Tuple, Type, TypeVar, Union)
 
+import sniffio
 import trio.from_thread
-from outcome import Error, Value
+from outcome import Error, Outcome, Value
+from trio.socket import SocketType as TrioSocketType
 from trio.to_thread import run_sync
 
 from .. import CapacityLimiterStatistics, EventStatistics, TaskInfo, abc
@@ -35,6 +39,11 @@ except ImportError:
 else:
     from trio.lowlevel import wait_readable, wait_writable
 
+try:
+    from trio.lowlevel import open_process as trio_open_process
+except ImportError:
+    from trio import open_process as trio_open_process
+
 T_Retval = TypeVar('T_Retval')
 T_SockAddr = TypeVar('T_SockAddr', str, IPSockAddrType)
 
@@ -60,17 +69,20 @@ sleep = trio.sleep
 #
 
 class CancelScope(BaseCancelScope):
-    def __new__(cls, original: Optional[trio.CancelScope] = None, **kwargs):
+    def __new__(cls, original: Optional[trio.CancelScope] = None,
+                **kwargs: object) -> 'CancelScope':
         return object.__new__(cls)
 
-    def __init__(self, original: Optional[trio.CancelScope] = None, **kwargs):
+    def __init__(self, original: Optional[trio.CancelScope] = None, **kwargs: object) -> None:
         self.__original = original or trio.CancelScope(**kwargs)
 
-    def __enter__(self):
+    def __enter__(self) -> 'CancelScope':
         self.__original.__enter__()
         return self
 
-    def __exit__(self, exc_type, exc_val, exc_tb):
+    def __exit__(self, exc_type: Optional[Type[BaseException]],
+                 exc_val: Optional[BaseException],
+                 exc_tb: Optional[TracebackType]) -> Optional[bool]:
         return self.__original.__exit__(exc_type, exc_val, exc_tb)
 
     def cancel(self) -> DeprecatedAwaitable:
@@ -93,6 +105,10 @@ class CancelScope(BaseCancelScope):
     def shield(self) -> bool:
         return self.__original.shield
 
+    @shield.setter
+    def shield(self, value: bool) -> None:
+        self.__original.shield = value
+
 
 CancelledError = trio.Cancelled
 checkpoint = trio.lowlevel.checkpoint
@@ -111,12 +127,12 @@ class ExceptionGroup(BaseExceptionGroup, trio.MultiError):
 
 
 class TaskGroup(abc.TaskGroup):
-    def __init__(self):
+    def __init__(self) -> None:
         self._active = False
         self._nursery_manager = trio.open_nursery()
-        self.cancel_scope = None
+        self.cancel_scope = None  # type: ignore[assignment]
 
-    async def __aenter__(self):
+    async def __aenter__(self) -> 'TaskGroup':
         self._active = True
         self._nursery = await self._nursery_manager.__aenter__()
         self.cancel_scope = CancelScope(self._nursery.cancel_scope)
@@ -132,13 +148,14 @@ class TaskGroup(abc.TaskGroup):
         finally:
             self._active = False
 
-    def start_soon(self, func: Callable, *args, name=None) -> None:
+    def start_soon(self, func: Callable, *args: object, name: object = None) -> None:
         if not self._active:
             raise RuntimeError('This task group is not active; no new tasks can be started.')
 
         self._nursery.start_soon(func, *args, name=name)
 
-    async def start(self, func: Callable[..., Coroutine], *args, name=None):
+    async def start(self, func: Callable[..., Coroutine],
+                    *args: object, name: object = None) -> object:
         if not self._active:
             raise RuntimeError('This task group is not active; no new tasks can be started.')
 
@@ -150,31 +167,59 @@ class TaskGroup(abc.TaskGroup):
 
 
 async def run_sync_in_worker_thread(
-        func: Callable[..., T_Retval], *args, cancellable: bool = False,
+        func: Callable[..., T_Retval], *args: object, cancellable: bool = False,
         limiter: Optional[trio.CapacityLimiter] = None) -> T_Retval:
-    def wrapper():
+    def wrapper() -> T_Retval:
         with claim_worker_thread('trio'):
             return func(*args)
 
-    return await run_sync(wrapper, cancellable=cancellable, limiter=limiter)
+    # TODO: remove explicit context copying when trio 0.20 is the minimum requirement
+    context = copy_context()
+    context.run(sniffio.current_async_library_cvar.set, None)
+    return await run_sync(context.run, wrapper, cancellable=cancellable, limiter=limiter)
 
-run_async_from_thread = trio.from_thread.run
-run_sync_from_thread = trio.from_thread.run_sync
+
+# TODO: remove this workaround when trio 0.20 is the minimum requirement
+def run_async_from_thread(fn: Callable[..., Awaitable[T_Retval]], *args: Any) -> T_Retval:
+    async def wrapper() -> Optional[T_Retval]:
+        retval: T_Retval
+
+        async def inner() -> None:
+            nonlocal retval
+            __tracebackhide__ = True
+            retval = await fn(*args)
+
+        async with trio.open_nursery() as n:
+            context.run(n.start_soon, inner)
+
+        __tracebackhide__ = True
+        return retval
+
+    context = copy_context()
+    context.run(sniffio.current_async_library_cvar.set, 'trio')
+    return trio.from_thread.run(wrapper)
+
+
+def run_sync_from_thread(fn: Callable[..., T_Retval], *args: Any) -> T_Retval:
+    # TODO: remove explicit context copying when trio 0.20 is the minimum requirement
+    return trio.from_thread.run_sync(copy_context().run, fn, *args)
 
 
 class BlockingPortal(abc.BlockingPortal):
-    def __new__(cls):
+    def __new__(cls) -> 'BlockingPortal':
         return object.__new__(cls)
 
-    def __init__(self):
+    def __init__(self) -> None:
         super().__init__()
         self._token = trio.lowlevel.current_trio_token()
 
     def _spawn_task_from_thread(self, func: Callable, args: tuple, kwargs: Dict[str, Any],
-                                name, future: Future) -> None:
+                                name: object, future: Future) -> None:
+        context = copy_context()
+        context.run(sniffio.current_async_library_cvar.set, 'trio')
         return trio.from_thread.run_sync(
-            partial(self._task_group.start_soon, name=name), self._call_func, func, args, kwargs,
-            future, trio_token=self._token)
+            context.run, partial(self._task_group.start_soon, name=name), self._call_func, func,
+            args, kwargs, future, trio_token=self._token)
 
 
 #
@@ -268,9 +313,12 @@ class Process(abc.Process):
         return self._stderr
 
 
-async def open_process(command, *, shell: bool, stdin: int, stdout: int, stderr: int):
-    process = await trio.open_process(command, stdin=stdin, stdout=stdout, stderr=stderr,
-                                      shell=shell)
+async def open_process(command: Union[str, Sequence[str]], *, shell: bool,
+                       stdin: int, stdout: int, stderr: int,
+                       cwd: Union[str, bytes, PathLike, None] = None,
+                       env: Optional[Mapping[str, str]] = None) -> Process:
+    process = await trio_open_process(command, stdin=stdin, stdout=stdout, stderr=stderr,
+                                      shell=shell, cwd=cwd, env=env)
     stdin_stream = SendStreamWrapper(process.stdin) if process.stdin else None
     stdout_stream = ReceiveStreamWrapper(process.stdout) if process.stdout else None
     stderr_stream = ReceiveStreamWrapper(process.stderr) if process.stderr else None
@@ -278,7 +326,7 @@ async def open_process(command, *, shell: bool, stdin: int, stdout: int, stderr:
 
 
 class _ProcessPoolShutdownInstrument(trio.abc.Instrument):
-    def after_run(self):
+    def after_run(self) -> None:
         super().after_run()
 
 
@@ -309,7 +357,7 @@ def setup_process_pool_exit_at_shutdown(workers: Set[Process]) -> None:
 #
 
 class _TrioSocketMixin(Generic[T_SockAddr]):
-    def __init__(self, trio_socket):
+    def __init__(self, trio_socket: TrioSocketType) -> None:
         self._trio_socket = trio_socket
         self._closed = False
 
@@ -340,7 +388,7 @@ class _TrioSocketMixin(Generic[T_SockAddr]):
 
 
 class SocketStream(_TrioSocketMixin, abc.SocketStream):
-    def __init__(self, trio_socket):
+    def __init__(self, trio_socket: TrioSocketType) -> None:
         super().__init__(trio_socket)
         self._receive_guard = ResourceGuard('reading from')
         self._send_guard = ResourceGuard('writing to')
@@ -429,7 +477,7 @@ class UNIXSocketStream(SocketStream, abc.UNIXSocketStream):
 
 
 class TCPSocketListener(_TrioSocketMixin, abc.SocketListener):
-    def __init__(self, raw_socket: socket.SocketType):
+    def __init__(self, raw_socket: socket.socket):
         super().__init__(trio.socket.from_stdlib_socket(raw_socket))
         self._accept_guard = ResourceGuard('accepting connections from')
 
@@ -445,7 +493,7 @@ class TCPSocketListener(_TrioSocketMixin, abc.SocketListener):
 
 
 class UNIXSocketListener(_TrioSocketMixin, abc.SocketListener):
-    def __init__(self, raw_socket: socket.SocketType):
+    def __init__(self, raw_socket: socket.socket):
         super().__init__(trio.socket.from_stdlib_socket(raw_socket))
         self._accept_guard = ResourceGuard('accepting connections from')
 
@@ -460,7 +508,7 @@ class UNIXSocketListener(_TrioSocketMixin, abc.SocketListener):
 
 
 class UDPSocket(_TrioSocketMixin[IPSockAddrType], abc.UDPSocket):
-    def __init__(self, trio_socket):
+    def __init__(self, trio_socket: TrioSocketType) -> None:
         super().__init__(trio_socket)
         self._receive_guard = ResourceGuard('reading from')
         self._send_guard = ResourceGuard('writing to')
@@ -482,7 +530,7 @@ class UDPSocket(_TrioSocketMixin[IPSockAddrType], abc.UDPSocket):
 
 
 class ConnectedUDPSocket(_TrioSocketMixin[IPSockAddrType], abc.ConnectedUDPSocket):
-    def __init__(self, trio_socket):
+    def __init__(self, trio_socket: TrioSocketType) -> None:
         super().__init__(trio_socket)
         self._receive_guard = ResourceGuard('reading from')
         self._send_guard = ResourceGuard('writing to')
@@ -555,7 +603,7 @@ getaddrinfo = trio.socket.getaddrinfo
 getnameinfo = trio.socket.getnameinfo
 
 
-async def wait_socket_readable(sock):
+async def wait_socket_readable(sock: socket.socket) -> None:
     try:
         await wait_readable(sock)
     except trio.ClosedResourceError as exc:
@@ -564,7 +612,7 @@ async def wait_socket_readable(sock):
         raise BusyResourceError('reading from') from None
 
 
-async def wait_socket_writable(sock):
+async def wait_socket_writable(sock: socket.socket) -> None:
     try:
         await wait_writable(sock)
     except trio.ClosedResourceError as exc:
@@ -578,34 +626,34 @@ async def wait_socket_writable(sock):
 #
 
 class Event(BaseEvent):
-    def __new__(cls):
+    def __new__(cls) -> 'Event':
         return object.__new__(cls)
 
-    def __init__(self):
+    def __init__(self) -> None:
         self.__original = trio.Event()
 
     def is_set(self) -> bool:
         return self.__original.is_set()
 
-    async def wait(self) -> bool:
+    async def wait(self) -> None:
         return await self.__original.wait()
 
     def statistics(self) -> EventStatistics:
         return self.__original.statistics()
 
-    def set(self):
+    def set(self) -> DeprecatedAwaitable:
         self.__original.set()
         return DeprecatedAwaitable(self.set)
 
 
 class CapacityLimiter(BaseCapacityLimiter):
-    def __new__(cls, *args, **kwargs):
+    def __new__(cls, *args: object, **kwargs: object) -> "CapacityLimiter":
         return object.__new__(cls)
 
-    def __init__(self, *args, original: Optional[trio.CapacityLimiter] = None):
+    def __init__(self, *args: object, original: Optional[trio.CapacityLimiter] = None) -> None:
         self.__original = original or trio.CapacityLimiter(*args)
 
-    async def __aenter__(self):
+    async def __aenter__(self) -> None:
         return await self.__original.__aenter__()
 
     async def __aexit__(self, exc_type: Optional[Type[BaseException]],
@@ -629,24 +677,24 @@ class CapacityLimiter(BaseCapacityLimiter):
     def available_tokens(self) -> float:
         return self.__original.available_tokens
 
-    def acquire_nowait(self):
+    def acquire_nowait(self) -> DeprecatedAwaitable:
         self.__original.acquire_nowait()
         return DeprecatedAwaitable(self.acquire_nowait)
 
-    def acquire_on_behalf_of_nowait(self, borrower):
+    def acquire_on_behalf_of_nowait(self, borrower: object) -> DeprecatedAwaitable:
         self.__original.acquire_on_behalf_of_nowait(borrower)
         return DeprecatedAwaitable(self.acquire_on_behalf_of_nowait)
 
     async def acquire(self) -> None:
         await self.__original.acquire()
 
-    async def acquire_on_behalf_of(self, borrower) -> None:
+    async def acquire_on_behalf_of(self, borrower: object) -> None:
         await self.__original.acquire_on_behalf_of(borrower)
 
     def release(self) -> None:
         return self.__original.release()
 
-    def release_on_behalf_of(self, borrower) -> None:
+    def release_on_behalf_of(self, borrower: object) -> None:
         return self.__original.release_on_behalf_of(borrower)
 
     def statistics(self) -> CapacityLimiterStatistics:
@@ -669,18 +717,20 @@ def current_default_thread_limiter() -> CapacityLimiter:
 # Signal handling
 #
 
-class _SignalReceiver(DeprecatedAsyncContextManager):
-    def __init__(self, cm):
+class _SignalReceiver(DeprecatedAsyncContextManager[T]):
+    def __init__(self, cm: ContextManager[T]):
         self._cm = cm
 
     def __enter__(self) -> T:
         return self._cm.__enter__()
 
-    def __exit__(self, exc_type, exc_val, exc_tb):
+    def __exit__(self, exc_type: Optional[Type[BaseException]],
+                 exc_val: Optional[BaseException],
+                 exc_tb: Optional[TracebackType]) -> Optional[bool]:
         return self._cm.__exit__(exc_type, exc_val, exc_tb)
 
 
-def open_signal_receiver(*signals: int):
+def open_signal_receiver(*signals: int) -> _SignalReceiver:
     cm = trio.open_signal_receiver(*signals)
     return _SignalReceiver(cm)
 
@@ -716,18 +766,18 @@ def get_running_tasks() -> List[TaskInfo]:
     return task_infos
 
 
-def wait_all_tasks_blocked():
+def wait_all_tasks_blocked() -> Awaitable[None]:
     import trio.testing
     return trio.testing.wait_all_tasks_blocked()
 
 
 class TestRunner(abc.TestRunner):
-    def __init__(self, **options):
+    def __init__(self, **options: object) -> None:
         from collections import deque
         from queue import Queue
 
-        self._call_queue = Queue()
-        self._result_queue = deque()
+        self._call_queue: "Queue[Callable[..., object]]" = Queue()
+        self._result_queue: Deque[Outcome] = deque()
         self._stop_event: Optional[trio.Event] = None
         self._nursery: Optional[trio.Nursery] = None
         self._options = options
@@ -737,7 +787,8 @@ class TestRunner(abc.TestRunner):
         async with trio.open_nursery() as self._nursery:
             await self._stop_event.wait()
 
-    async def _call_func(self, func, args, kwargs):
+    async def _call_func(self, func: Callable[..., Awaitable[object]],
+                         args: tuple, kwargs: dict) -> None:
         try:
             retval = await func(*args, **kwargs)
         except BaseException as exc:
@@ -745,7 +796,7 @@ class TestRunner(abc.TestRunner):
         else:
             self._result_queue.append(Value(retval))
 
-    def _main_task_finished(self, outcome) -> None:
+    def _main_task_finished(self, outcome: object) -> None:
         self._nursery = None
 
     def close(self) -> None:
@@ -754,7 +805,8 @@ class TestRunner(abc.TestRunner):
             while self._nursery is not None:
                 self._call_queue.get()()
 
-    def call(self, func: Callable[..., Awaitable], *args, **kwargs):
+    def call(self, func: Callable[..., Awaitable[T_Retval]],
+             *args: object, **kwargs: object) -> T_Retval:
         if self._nursery is None:
             trio.lowlevel.start_guest_run(
                 self._trio_main, run_sync_soon_threadsafe=self._call_queue.put,
